from dataclasses import dataclass, field
from html.parser import HTMLParser
from pathlib import Path
from textwrap import indent
from typing import List, Optional

from nbconvert import HTMLExporter
from nbconvert.preprocessors import tagremove

from . import preprocessors, utils


@dataclass
class TOCTree:
    title: Optional[str] = None
    anchor: Optional[str] = None
    children: List["TOCTree"] = field(default_factory=list)

    def __str__(self):
        if self.children:
            return (self.title or "<root>") + "\n" + indent("\n".join(str(c) for c in self.children), " ")
        else:
            return self.title

    def to_html_list(self, url="", max_depth=1, wrap=False, depth=0) -> str:
        anchor_text = f"#{self.anchor}" if self.anchor else ""
        link_text = f'<a href="{url}{anchor_text}">{self.title}</a>' if self.anchor is not None else self.title

        if self.children and depth < max_depth:
            children = "\n".join(c.to_html_list(url, max_depth, depth=depth + 1) for c in self.children)
            if self.title:
                return f"<li>{link_text}\n <ol>\n{indent(children, '  ')}\n </ol>\n</li>"
            else:
                if wrap:
                    return f"<ol>\n{indent(children, ' ')}\n</ol>"
                else:
                    return indent(children, ' ')
        else:
            if self.title:
                return f"<li>{link_text}</li>"
            else:
                return ""


class MyHTMLParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.in_header = False
        self.current_header_level = None
        self.current_header_title = None
        self.current_header_link = None
        self.headers = []

    def handle_starttag(self, tag, attrs):
        if tag in {"h1", "h2", "h3"}:
            attrs = dict(attrs)
            if attrs.get("id") in {"main-title", "Exercise", "Exercises"}:
                return
            self.in_header = True
            self.current_header_level = tag[1]
            self.current_header_link = attrs.get("id")

    def handle_endtag(self, tag):
        if tag in {"h1", "h2", "h3"} and self.in_header:
            self.headers.append({"level": int(self.current_header_level), "title":  self.current_header_title, "anchor": self.current_header_link})

            self.in_header = False
            self.current_header_level = None
            self.current_header_title = None
            self.current_header_link = None

    def handle_data(self, data):
        if self.in_header and self.current_header_title is None:
            self.current_header_title = data


def get_headers(html: str) -> TOCTree:
    """
    Given a HTML string, extract the table of contents
    """
    parser = MyHTMLParser()
    parser.feed(html)
    parser.close()

    toc = TOCTree()

    for header in parser.headers:
        temp_toc = toc
        new_toc = TOCTree(header["title"], header["anchor"])
        for _ in range(header["level"] - 1):
            try:
                temp_toc = temp_toc.children[-1]
            except IndexError:
                pass
        temp_toc.children.append(new_toc)

    if len(toc.children) == 1:
        toc.children[0].anchor = ""

    return toc


def construct_toc(chapters: List[Path], config, skip_first_page=True) -> str:
    """
    Given some ipynb filenames, return the HTML table of contents
    """
    toc = []

    plain_html_exporter = HTMLExporter(config=config)
    plain_html_exporter.register_preprocessor(tagremove.TagRemovePreprocessor, enabled=True)
    plain_html_exporter.register_preprocessor(preprocessors.PageLinks(chapters), enabled=True)

    for filename in chapters:
        body, resources = utils.ipynb_to_html(plain_html_exporter, filename)
        out_filename = resources["output_stem"] + resources["output_extension"]
        header_list = get_headers(body).to_html_list(out_filename, max_depth=1)
        toc.append(header_list)

    if skip_first_page:
        toc = toc[1:]

    inner_toc_string = '\n'.join(toc)
    full_toc_html = f"<ol>\n{inner_toc_string}\n</ol>"
    return full_toc_html



import base64
import hashlib
import re
import shlex
from pathlib import Path

from nbconvert.preprocessors import Preprocessor


class HighlightExercises(Preprocessor):
    """
    Any cells with the tag ``exercise`` will have the ``exercise`` css class attached to them.
    By default this highlights the output cell in yellow.
    """

    def preprocess(self, notebook, resources):
        exercise_num = 1
        for cell in notebook.cells:
            if "tags" in cell["metadata"]:
                if "exercise" in cell["metadata"]["tags"]:
                    cell.custom_class = "exercise"

                    if cell.source.startswith("### Exercise #"):
                        cell.source = cell.source.replace("### Exercise #", f"### Exercise {exercise_num}")
                        exercise_num += 1
        return notebook, resources


class PageLinks(Preprocessor):
    """
    Adds in links from one page to the next.
    The order is based on the naming of the main pages' numbers.
    i.e. ``00 Intro.ipynb` will have a next link to ``01 Foo.ipynb``.
    """

    def __init__(self, chapters, **kw):
        super().__init__(**kw)
        self.chapters = chapters

    def preprocess(self, notebook, resources):
        source_file = notebook["metadata"]["nbpretty"]["source_file"]
        files = sorted(f.stem for f in self.chapters)

        try:
            index = files.index(source_file.stem)
        except ValueError:
            # Not a chapter, i.e. an answer page or an aside
            resources["output_stem"] = source_file.stem
            return notebook, resources

        output_files = [f[3:] for f in files]  # Remove the 01, 02, 03 etc.
        output_files[0] = "index"
        resources["output_stem"] = output_files[index]

        if len(files) < 2:
            return notebook, resources

        if index == 0:
            previous_file = None
            next_file = output_files[1]
        elif index == len(output_files) - 1:
            previous_file = output_files[-2]
            next_file = None
        else:
            previous_file = output_files[index - 1]
            next_file = output_files[index + 1]

        if index == 1:
            previous_file = "index"

        source = ""
        if previous_file is not None:
            source += f'[<font size=\"5\">Previous</font>]({previous_file}.html)'
        if previous_file is not None and next_file is not None:
            source += '<font size=\"5\"> | </font>'
        if next_file is not None:
            source += f'[<font size=\"5\">Next</font>]({next_file}.html)'

        notebook.cells.append({"cell_type": "markdown", "metadata": {}, "source": source})

        return notebook, resources


class SetTitle(Preprocessor):
    """
    Based on the ``course_title`` in ``config.yaml``
    this sets the title at the top of each page and in the HTML header.
    """

    def __init__(self, course_title, **kw):
        super().__init__(**kw)
        self.course_title = course_title

    def preprocess(self, notebook, resources):
        notebook["metadata"]["course_title"] = self.course_title
        notebook["metadata"]["title"] = f"{notebook['metadata']['chapter_title']} - {self.course_title}"
        return notebook, resources


class HideWriteFileMagic(Preprocessor):
    """
    This looks for cells which start with ``%%writefile``, ``%run``, ``%cd`` or ``!`` and hides that from the output.
    For ``%%writefile`` it removes that line, sets the output prompt to the name of the file and hides the output "Overwriting foo.py".
    For ``%run foo.py`` it turns it into ``python foo.py`` and sets the prompt to ``$`` to indicate running in the terminal.
    For ``%cd`` it changes it into ``cd`` and for ``!`` it just shows the run command.
    """

    def preprocess(self, notebook, resources):
        execution_count = 0
        for cell in notebook.cells:
            # Get any potential cell execute results
            execute_results = [o for o in cell.get("outputs", []) if o.get("output_type") == "execute_result"]

            if cell["source"].startswith("%%writefile"):
                #TODO Give better error message for cell which just contains `%%writefile` and nothing else.
                file_name = cell["source"].split("\n")[0].split()[1]
                cell.metadata["writefile"] = file_name

                to_remove = len(cell["source"].split("\n")[0])
                cell["source"] = cell["source"][to_remove:]

                if Path(file_name).suffix == ".py":
                    cell.metadata["magics_language"] = "python"
                else:
                    cell.metadata["magics_language"] = "text"

                cell["outputs"] = []
            elif any(execute_results) and "text/x.prog" in execute_results[0]["metadata"]:
                cell.metadata["runcommand"] = ""
                cell.metadata["magics_language"] = "bash"
                command = execute_results[0]["metadata"]["text/x.prog"]["command"]
                if cell["source"].startswith("%run_python_script") or cell["source"].startswith("%%run_python_script"):
                    commands = shlex.split(command)
                    commands[0] = "python"
                    command = shlex.join(commands)
                cell["source"] = command

                cell["outputs"][0] = {"name": "stdout", "output_type": "stream", "text": execute_results[0]["data"]["text/plain"]}
            elif cell["source"].startswith("%run"):
                script = cell["source"][5:]
                cell.metadata["runcommand"] = ""
                cell.metadata["magics_language"] = "bash"

                cell["source"] = f"python {script}"
            elif any(l.startswith("!$sys.executable") for l in cell["source"].split("\n")):
                command = re.sub(r".*(!\$sys\.executable)(.*)", r"python\g<2>", cell["source"], flags=re.DOTALL)
                cell.metadata["runcommand"] = ""
                cell.metadata["magics_language"] = "bash"

                cell["source"] = command
            elif cell["source"].startswith("!"):
                command = cell["source"][1:]
                if "COLUMNS=" in command:
                    command = command[len("COLUMNS=nn")+1:]
                if "venv/bin/" in command:
                    command = command[9:]
                cell.metadata["runcommand"] = ""
                cell.metadata["magics_language"] = "bash"

                cell["source"] = command
            elif cell["source"].startswith("%cd"):
                file_name = cell["source"].split()[1]
                cell.metadata["runcommand"] = ""
                cell.metadata["magics_language"] = "bash"
                cell["source"] = f"cd {file_name}"
                cell["outputs"] = []
            else:
                if "execution_count" in cell:
                    execution_count += 1
                    cell.execution_count = execution_count
        return notebook, resources


class FixLinkExtensions(Preprocessor):
    """
    Replaces all ``.ipynb`` with ``.html`` so that links continue to work.
    """

    def preprocess(self, notebook, resources):
        for cell in notebook.cells:
            # TODO Only change links to local files
            cell["source"] = cell["source"].replace(".ipynb", ".html")
        return notebook, resources


class CustomBlocks(Preprocessor):
    """
    Based on the ``custom_blocks`` list in ``config.yaml``
    it sets all cells with a tag matching a custom block name
    to be collapsible by clicking on the image that goes in the prompt.
    """

    def __init__(self, custom_blocks, **kw):
        super().__init__(**kw)
        self.custom_blocks = custom_blocks

    def preprocess(self, notebook, resources):
        "dsdf"
        if not self.custom_blocks:
            return notebook, resources

        for i, cell in enumerate(notebook.cells):
            for tag, block in self.custom_blocks.items():
                if "tags" in cell["metadata"] and tag in cell["metadata"]["tags"]:
                    cell.metadata["custom_block"] = block
                    cell.metadata["custom_block"]["id"] = i
                    cell.metadata["custom_block"]["name"] = tag
                    cell.custom_class = "outline"

        return notebook, resources


class InsertTOC(Preprocessor):
    """
    Replaces all cells with the tag ``toc`` with the table on contents,
    based on the file naming.
    """

    def __init__(self, toc_html, **kw):
        super().__init__(**kw)
        self.toc_html = toc_html

    def preprocess_cell(self, cell, resources, index):
        if "tags" in cell["metadata"] and "toc" in cell["metadata"]["tags"]:
            cell.source = self.toc_html
        return cell, resources


class UninlineCss(Preprocessor):
    """
    Moves the inlined CSS generated by nbconvert into a separate file
    which is referenced by all notebooks.
    """

    def preprocess(self, notebook, resources):
        def import_css(name):
            import jinja2
            env = self.parent.environment
            css_source = env.loader.get_source(env, name)[0]
            h = base64.urlsafe_b64encode(hashlib.sha1(css_source.encode()).digest()[:12]).decode()
            # static/style.css becomes, e.g. nbpretty_oNmeLFPs3BZEPiaf_style.css
            filename = Path(f"nbpretty_{h}_{Path(name).stem}.css")
            resources["files"][filename] = css_source
            code = f"""<style type="text/css">\n@import '{filename}'\n</style>"""
            return jinja2.Markup(code)

        resources['import_css'] = import_css
        resources['include_css'] = import_css

        if "files" not in resources:
            resources["files"] = {}
        if "inlining" not in resources:
            return notebook, resources
        for i, css in enumerate(resources["inlining"]["css"]):
            h = base64.urlsafe_b64encode(hashlib.sha1(css.encode()).digest()[:12]).decode()
            filename = Path(f"nbpretty_{h}.css")
            resources["inlining"]["css"][i] = f"@import '{filename}'"
            resources["files"][filename] = css
        return notebook, resources


class AddExtraCss(Preprocessor):
    """
    If the extra CSS file is give, add it to the metadata.
    """

    def __init__(self, extra_css_file, **kw):
        super().__init__(**kw)
        self.extra_css_file = extra_css_file

    def preprocess(self, notebook, resources):
        notebook["metadata"]["extra_css_file"] = self.extra_css_file
        return notebook, resources



from functools import partial
import importlib.metadata
import logging
import os
import re
import shutil
from pathlib import Path
from typing import Dict

import click
from livereload import Server
from rich.logging import RichHandler
from rich.progress import track
from yaml import safe_load

from nbconvert import HTMLExporter
from nbconvert.preprocessors import tagremove
from traitlets.config import Config

from . import preprocessors, toc, utils

__version__ = importlib.metadata.version(__package__)


@click.command()
@click.argument("directory", type=Path)
@click.option('--prompt/--no-prompt', default=True, help="Should the output contain the IPython prompt")
@click.option('--serve', is_flag=True, help="Serve the pages to the browser")
def build(directory: Path, prompt: bool, serve: bool):
    """
    Convert notebooks in DIRECTORY
    """

    if not directory.exists() or not directory.is_dir():
        ctx = click.get_current_context()
        ctx.fail(f"Directory {directory.resolve()} does not exist")

    _build(directory, prompt)

    if serve:
        server = Server()
        server.watch(f"{str(directory)}/*.ipynb", partial(_build, directory, prompt))
        server._setup_logging = lambda: None  # https://github.com/lepture/python-livereload/issues/232
        server.serve(root=directory, open_url_delay=1)


def _build(directory: Path, prompt: bool):
    logging.basicConfig(level=logging.INFO, format="%(message)s", handlers=[RichHandler(show_time=False)])
    logger = logging.getLogger("nbpretty")

    mod_dir = Path(os.path.abspath(os.path.dirname(__file__)))

    config = Config()
    config.TagRemovePreprocessor.remove_cell_tags = {"remove_cell"}
    config.TagRemovePreprocessor.remove_input_tags = {"remove_input"}
    config.TemplateExporter.exclude_input_prompt = not prompt
    config.TemplateExporter.exclude_output_prompt = not prompt
    config.TemplateExporter.extra_template_basedirs = [mod_dir / "templates"]
    config.Exporter.optimistic_validation = True

    chapters = sorted(f for f in directory.glob("*.ipynb") if re.match(r"\d\d.*", f.name))
    answers = sorted(f for f in directory.glob("answer_*.ipynb"))
    asides = sorted(f for f in directory.glob("aside_*.ipynb"))
    appendixes = sorted(f for f in directory.glob("appendix_*.ipynb"))

    try:
        with (directory / "config.yaml").open() as f:
            nbpretty_config = safe_load(f)
            course_title = nbpretty_config["course_title"]
            custom_blocks = nbpretty_config.get("custom_blocks")
            try:
                extra_css_file = nbpretty_config["extra_css_file"]
            except KeyError:
                extra_css_file = None
    except FileNotFoundError:
        ctx = click.get_current_context()
        ctx.fail(f"config.yaml could not be found")

    logger.info("Constructing table of contents")
    toc_html = toc.construct_toc(chapters, config)

    html_exporter = HTMLExporter(template_name="nbpretty", config=config)
    html_exporter.register_preprocessor(tagremove.TagRemovePreprocessor, enabled=True)
    html_exporter.register_preprocessor(preprocessors.PageLinks(chapters), enabled=True)
    html_exporter.register_preprocessor(preprocessors.HighlightExercises, enabled=True)
    html_exporter.register_preprocessor(preprocessors.SetTitle(course_title), enabled=True)
    html_exporter.register_preprocessor(preprocessors.HideWriteFileMagic, enabled=True)
    html_exporter.register_preprocessor(preprocessors.FixLinkExtensions, enabled=True)
    html_exporter.register_preprocessor(preprocessors.CustomBlocks(custom_blocks), enabled=True)
    html_exporter.register_preprocessor(preprocessors.InsertTOC(toc_html), enabled=True)
    html_exporter.register_preprocessor(preprocessors.UninlineCss, enabled=True)
    html_exporter.register_preprocessor(preprocessors.AddExtraCss(extra_css_file), enabled=extra_css_file is not None)

    output_directory = directory

    extra_files: Dict[Path, str] = {}

    all_files = chapters + answers + asides + appendixes

    for filename in track(all_files):
        body, resources = utils.ipynb_to_html(html_exporter, filename)
        extra_files.update(resources["files"])
        with open(output_directory / f"{resources['output_stem']}.html", "w") as out:
            logger.info(f"Writing '{filename}' as '{out.name}'")
            out.write(body)

    for filename, contents in extra_files.items():
        with (output_directory / filename).open("w") as f:
            f.write(contents)


def main():
    build()


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Tuple, Any

import nbformat
from nbconvert import HTMLExporter


def ipynb_to_html(html_exporter: HTMLExporter, filename: Path) -> Tuple[str, Any]:
    """
    Convert a ipynb file to HTML text using an exporter
    """
    chapter_title = filename.name.lstrip("1234567890").lstrip(" ")[:-6]
    with filename.open() as infile:
        notebook = nbformat.read(infile, as_version=4)
        notebook["metadata"]["nbpretty"] = {"source_file": filename}
        notebook["metadata"]["chapter_title"] = chapter_title
        return html_exporter.from_notebook_node(notebook)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nbpretty']

package_data = \
{'': ['*'], 'nbpretty': ['templates/nbpretty/*', 'templates/nbpretty/static/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.0,<9.0',
 'ipython>=8.0.0,<9.0.0',
 'jinja2<3.1',
 'livereload>=2.0.0,<3.0.0',
 'nbconvert>=6.4.0,<7.0.0',
 'rich>=12.0.0,<13.0.0']

entry_points = \
{'console_scripts': ['nbpretty = nbpretty:main']}

setup_kwargs = {
    'name': 'nbpretty',
    'version': '0.2.14',
    'description': 'A tool to convert sets of Jupyter notebook files into a single, cohesive set of linked pages',
    'long_description': 'nbpretty\n========\n\nnbpretty is a tool to convert sets of notebook files into a single, cohesive set of linked pages.\n\nDocumentation at https://nbpretty.readthedocs.io\n',
    'author': 'Matt Williams',
    'author_email': 'matt@milliams.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/milliams/nbpretty',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
